function [k,K] = BuildIntGram(x,A,S,opts)
% constructs an MxM matrix containing elements
% K_{kl} = lambda.^-2 sum_n S_kn S_ln int_0 ^1 int_0 ^1 k(x_k(t),x_l(t')) dt dt'
%
% and an NxM matrix containing elements
% k_l = lambda.^-2 S_ln int_0 ^1  k(x,x_l(t)) dt
%
% copyright (c) 2012 Philipp Hennig

%% precomputations 
% for very large N, this is the most expensive part of this function
[M,N] = size(A); lambda = opts.lambda;

up = sqrt(sum(S.^2,2));
E  = bsxfun(@rdivide,S,up);
EA = E * A'; 
AA = A * A';
EE = E * E';
eLx = lambda.^2 .\ (E * x'); 
aLx = lambda.^2 .\ (A * x'); 
xLx = lambda.^2 .\ (x * x');

clear S A; % to save memory. We don't need any of the expensive bits after this

eLe = lambda.^2 .\ EE;
eLa = lambda.^2 .\ EA;
aLa = lambda.^2 .\ AA;
eVe = (opts.thetaH.^2 ./ opts.lambda0.^2) .* EE;  % MxM

%% build K
switch isfield(opts,'Kh')
    case true % re-use old Gram matrix
        K = opts.Kh;
        % the diagonal of K consists of double univariate integrals
        K(M,M) = (sqrt(2*pi) * lambda * (up(M).*erf(up(M)./(sqrt(2)*lambda))) ...
            + 2 * lambda.^2 * (exp(-up(M).^2 ./ (2*lambda.^2))  - 1)) * eVe(M,M);
        % the off-diagonals consist of single bivariate integrals
        for k = 1:M-1
            detA = eLe(k,k) * eLe(M,M) - eLe(k,M) * eLe(k,M);
            ai = [eLe(k,k),  eLe(k,M);  eLe(M,k), eLe(M,M)] ./ detA;
            b  = [eLa(M,M) - eLa(M,k); eLa(k,k) - eLa(k,M)];
            c  = aLa(M,M) + aLa(k,k) - aLa(k,M) - aLa(M,k);
            rho= eLe(k,M) / sqrt(eLe(k,k) * eLe(M,M));
            %if abs(rho) - 1 < 1.0e-14; rho = sign(rho) * (1-1.0e-14); end;
            ui = sqrt(1 - rho*rho) * diag(sqrt([eLe(M,M),eLe(k,k)])) * ai * b;
            uf = (sqrt(1 - rho*rho) * diag(sqrt([eLe(M,M),eLe(k,k)])) ...
                * [up(M);up(k)]) + ui;
            et = exp(-0.5 * (c-b' * ai * b));
            preterm= 2 * pi / sqrt(detA);
            K(k,M) = eVe(k,M) * preterm * et * (BivariateNormalCDF(uf(1),uf(2),rho) ...
                -BivariateNormalCDF(uf(1),ui(2),rho) ...
                -BivariateNormalCDF(ui(1),uf(2),rho) ...
                +BivariateNormalCDF(ui(1),ui(2),rho));
            K(M,k) = K(k,M); % Gram matrix is symmetric
        end
    case false % if we build up the matrix from scratch
        if M > 1; warning('SCIENCE:DeprecationOfKnowledge','could not find Gram matrix; building from scratch.'); end;
        % the diagonal of K consists of double univariate integrals
        K = diag(sqrt(2*pi) * lambda * (up.*erf(up./(sqrt(2)*lambda))) ...
            + 2 * lambda.^2 * (exp(-up.^2 ./ (2*lambda.^2))  - 1));
        
        % the off-diagonals consist of single bivariate integrals
        for k = 1:M
            for l = k+1:M
                detA = eLe(k,k) * eLe(l,l) - eLe(k,l) * eLe(k,l);
                % a  = [eLe(l,l), -eLe(k,l); -eLe(k,l), eLe(k,k)];
                ai = [eLe(k,k),  eLe(k,l);  eLe(l,k), eLe(l,l)] ./ detA;
                b  = [eLa(l,l) - eLa(l,k); eLa(k,k) - eLa(k,l)];
                c  = aLa(l,l) + aLa(k,k) - aLa(k,l) - aLa(l,k);
                rho= eLe(k,l) / sqrt(eLe(k,k) * eLe(l,l));
                ui = sqrt(1 - rho*rho) * diag(sqrt([eLe(l,l),eLe(k,k)])) * ai * b;
                uf = (sqrt(1 - rho*rho) * diag(sqrt([eLe(l,l),eLe(k,k)])) ...
                    * [up(l);up(k)]) + ui;
                et = exp(-0.5 * (c-b' * ai * b));
                preterm= 2 * pi / sqrt(detA);
                K(k,l) = preterm * et * (BivariateNormalCDF(uf(1),uf(2),rho) ...
                    -BivariateNormalCDF(uf(1),ui(2),rho) ...
                    -BivariateNormalCDF(ui(1),uf(2),rho) ...
                    +BivariateNormalCDF(ui(1),ui(2),rho));
                K(l,k) = K(k,l); % Gram matrix is symmetric
            end
        end
        K = eVe .* K;
end

%% build k
a = sqrt(diag(eLe)); 
b = diag(eLa) - eLx; 
c = xLx - 2 * aLx + diag(aLa);

pf= exp(-0.5.*(c - b.^2 ./ a.^2)) .* sqrt(pi./(2.*a));
k = pf .* (erf((up.*a.^2 + b)./ a ./ sqrt(2)) - erf(b ./ a ./ sqrt(2)));
k = bsxfun(@times,(opts.thetaH.^2 ./ opts.lambda0.^2) .* E',k');