function [p,opts] = NoisyQuasiNewton(x,X,Y,opts)
% Quasi-Newton directions from noisy observations
%
%                                                   size
% inputs: current location         x                 1xN
%         past locations           X                 MxN
%         past observed gradients  Y                 MxN
%         options/hyperparameters  opts      with fields
%         kernel signal variance   opts.theta        1x1
%         kernel length scale      opts.lambda       1x1
%         prior mean gradient      opts.grad_mean    @(MxN) -> MxN
%         map vector v to B0 *v    opts.Hess_mean    @(MxN) -> MxN
%         map vector v to B0 \v    opts.invHess_mean @(MxN) -> MxN
%         observation noise        opts.sigma        1x1
%
% output: new search direction     p   an estimate of - B^{-1}(x) * grad(x)
%
% this work is under review by NIPS 2012. Do not distribute.
% copyright (c) 2012 Philipp Hennig. All rights reserved.
hyp = log([opts.lambda;opts.thetaG]);
[M,N] = size(X);

if M == 1; p = -opts.invHess_mean(Y); return; end;       % only one observation: gradient descent

%% construct estimate of gradient at x
mgX= opts.grad_mean(X);                           % past prior means       MxN
mgx= opts.grad_mean(x);                           % current prior mean     1xN
K  = feval(@covSEiso,hyp,X);                      % kernel Gram matrix     MxM
L  = opts.lambda0.^2 \ K + opts.sigma.^2 * eye(M);% full Gram matrix       MxM
k  = feval(@covSEiso,hyp,x,X);                    % kernel projection      1xM
Yc = Y - mgX;                                     % mean-adjusted data     MxN
gx = mgx + (opts.lambda0.^2 \ (k / L)) * Yc;      % post. mean gradient    1xN

%% construct estimate of inverse Hessian at x
S  = X(2:end,:) - X(1:end-1,:);
A  = X(1:end-1,:);                                % integration boundaries M-1xN
U  = Y(2:end,:) - Y(1:end-1,:);                   % integral observations  M-1xN
mHX= opts.Hess_mean(S);                           % past Hessian means     M-1xN
[k,K] = BuildIntGram(x,A,S,opts);                 % kernel Gram matrix   M-1xM-1
opts.Kh = K;                                      % store, for efficiency
                                                  % kernel projection      NxM-1
UmB   = (U - mHX);                                % estimate deviation   M-1xN
UBS   = UmB * S'; UBS = UBS + UBS';               % projected to M-space M-1xM-1
V     = k / (K + (2*opts.sigma.^2) * eye(M-1));   % projection to Hessian  NxM-1
W     = UmB' - 0.25 * (V * UBS);                  % projection to Hessian  NxM-1

%% multiply the two to get p
VH    = opts.invHess_mean(V');                    % linear complexity products
WH    = opts.invHess_mean(W');                    % with prior Hessian     MxN
% Schur complement 2(M-1)x2(M-1):
A = [VH * V, VH * W + eye(M-1); WH * V + eye(M-1), WH * W]; %+ 1.0e-8 * eye(2*(M-1)); 

p     = -opts.invHess_mean(gx) + ([VH', WH'] * (A \ [VH * gx'; WH * gx']))';
%% if used for experiments, also return diagnostics, precomputations
if isfield(opts,'debug') && opts.debug
    opts.alphaGrad = L \ Yc;
    opts.HessGram  = (K + (2*opts.sigma.^2) * eye(M-1));
    opts.HessUmB   = UmB;
end